﻿-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = AceLibrary("AceLocale-2.2"):new("Violation: Critical hit rate")

L:RegisterTranslations("enUS", function() return {
	["Critical hit rate"] = true,
	["Crit%"] = true,
} end)

L:RegisterTranslations("koKR", function() return {
	["Critical hit rate"] = "치명타율",
	["Crit%"] = "치명타",
} end)

-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------

local playerName = nil
local data = nil
local broadcastdirty = nil
local dirty = nil
local moduleName = "HitCritRate"
local hitCount = {}
local critCount = {}

local roster = AceLibrary("Roster-2.1")
local parser = ParserLib:GetInstance("1.1")

-----------------------------------------------------------------------
-- Module declaration
-----------------------------------------------------------------------

local mod = Violation:NewModule(moduleName)
mod.revision = tonumber(("$Revision: 38177 $"):sub(12, -3)) or 1
mod.version = 4
mod.dataFormatter = "%5.1f%%"
mod.hidePercentage = true

-----------------------------------------------------------------------
-- Module initialization
-----------------------------------------------------------------------

function mod:OnModuleRegister()
	playerName = UnitName("player")
	data = mod:new()
end

function mod:OnModuleEnable()
	parser:RegisterInfoType( moduleName, "hit", self.OnParserEvent)
end

function mod:OnModuleDisable()
	parser:UnregisterAllInfoTypes( moduleName )
end

function mod:OnReset()
	if type(data) == "table" then
		data = mod:del(data)
	end

	data = mod:new()
	data[playerName] = 0
	
	if type(hitCount) == "table" then
		hitCount = mod:del(hitCount)
	end
	hitCount = mod:new()
	
	if type(critCount) == "table" then
		critCount = mod:del(critCount)
	end
	critCount = mod:new()
	
	dirty = nil
	broadcastdirty = nil
end

-----------------------------------------------------------------------
-- Module implementation
-----------------------------------------------------------------------

function mod:GetRevision() return self.revision end
function mod:GetDisplayName() return L["Critical hit rate"] end
function mod:GetShortDisplayName() return L["Crit%"] end

function mod:GetBroadcastData()
	broadcastdirty = nil
	return data[playerName]
end

function mod:IsDirty()
	return dirty
end

function mod:ResetDirty()
	dirty = nil
end

function mod:IsBroadcastDirty()
	return broadcastdirty
end

function mod:OnRemoteData(sender, input)
	if input ~= data[sender] then
		data[sender] = input
		dirty = true
	end
end

function mod:OnParserEvent( info )
	local name, amount, victim = info.source, info.amount, info.victim

	if not name or name == ParserLib_NONE or not amount or amount == 0 or info.isDOT then
		return
	end

	if not Violation:AcceptDataFromSource(name, moduleName) then 
		return
	end
	
	-- We ignore split damage from known units to known units.
	-- this includes soullink damage from you to your pet.
	-- thanks rophy for pointing this out.
	if info.isSplit and victim then
		if victim == ParserLib_SELF or roster:GetUnitIDFromName( victim ) then
			return
		end
	end
	
	hitCount[name] = (hitCount[name] or 0) + 1
	if info.isCrit then
		critCount[name] = (critCount[name] or 0) + 1
	end
	
	local newCritRate =  math.floor(100.0 * (critCount[name] or 0) / (hitCount[name] or 1))
	if name == ParserLib_SELF then
		if newCritRate ~= data[playerName] then
			data[playerName], dirty, broadcastdirty = newCritRate, true, true
		end
	else
		mod:OnRemoteData(name, newCritRate)
	end
end

function mod:GetData(player)
	if player then
		return data[player]
	else
		return data
	end
end
