﻿-------------------------------------------------------------------
-- To do and/or ideas
--
-- From Rabbit:
-- * Parse damage messages from swstats and damagemeters.
-- * Automatically output our damage done and position to chat when a boss has
-- been killed; this is easy with Babble-Boss, CHAT_MSG_HOSTILE_DEATH and
-- UnitClassification.
-- * Maybe add an auto-reset for boss kill as well.
--
-- * Problem: Data for users without Violation will be different from Violation
-- user to Violation user because of combat log range.
--   Solution: At all times there is one person in the group with Violation
--   leader status - this status is first determined by who is the group leader
--   or officer, and then randomly. When a leader is selected, he checks what
--   other people in the group have Violation and assign them parsing targets -
--   for example if group 3 has 1 Violation user, that user will report for the
--   rest of the members in group 3. If noone in a group has Violation, a random
--   reporter of first the same class, then the same role, and then completely
--   random will be assigned to parse his data.
--
-- From Nymbia:
-- * Pick up the slack of users without the addon by splitting up the members
-- without the addon among those with it.  ie, in a 25 man raid where 10 users
-- have Violation, each of them is assigned to parse the damage of either 1 or
-- 2 other raid members.
--    Mikk: What's the gain of this? You still have to process everyone's events. 
--          And you keep stats for them in RAM anyway via sync.
--          Plus you get the problem of range.  Just process events of everyone not running vio/DM imo.
-- * Possibly keep two running totals:  'this fight' and 'overall', to allow
-- for statistics to be displayed for a single encounter without requiring
-- constant resets.
--
-- From Adirelle:
-- * Extend communication protocol to allow sending of pet data (at least).
-- * Multidisplay
-- * Character tooltip on hovering over bars
--
-- From <person>:
-- * Brew my coffee when I win the damage meter contest after a boss fight.
--
-----------------------------------------------------------------------

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = AceLibrary("AceLocale-2.2"):new("Violation")

L:RegisterTranslations("enUS", function() return {
	["Track"] = true,
	["What data types to track."] = true,
	["Tooltip"] = true,
	["What data to show in the tooltip."] = true,
	["Toggle whether to display data from %q in the tooltip."] = true,
	["Colors"] = true,
	["Assign colors to the different data types."] = true,
	["What color to use for the %q data type."] = true,
	["Color the tooltip"] = true,
	["Toggles whether or not to use these colors in the tooltip."] = true,
	
	["Bars"] = true,
	["Settings for the GUI."] = true,
	["Toggle the GUI."] = true,
	["Show"] = true,
	["Display"] = true,
	["What data to show in the GUI."] = true,
	["Texture"] = true,
	["Set the bar texture."] = true,
	["Scale"] = true,
	["Set the bar window scale."] = true,
	["Window Color"] = true,
	["Set the GUI Color"] = true,

	["Resetting data as requested by %s."] = true,
	["Reset"] = true,
	["Options for when to reset the data."] = true,
	["Reset now"] = true,
	["Immediately resets your data."] = true,
	["Reset group"] = true,
	["Resets the Violation data for every group member. Only the group leader or an officer can do this."] = true,
	["Leader resets"] = true,
	["Accept reset requests from raid leaders or officers."] = true,
	["Auto-reset"] = true,
	["Entering combat"] = true,
	["Automatically reset the data when entering combat."] = true,
	["Leaving combat"] = true,
	["Automatically reset the data when leaving combat."] = true,
	["Print on reset"] = true,
	["Prints your current data locally when your data is reset."] = true,

	["Text"] = true,
	["What data to show on the FuBar panel, if available."] = true,

	["Toggle whether to track %q."] = true,
	["Toggle whether to display data from %q on the FuBar panel."] = true,

	["%s has been disabled, since it is not compatible with the latest module API."] = true,
	["No modules running, please enable one or more from the configuration menu."] = true,
	["No data gathered yet."] = true,
	["|cffeda55fClick|r to toggle the bar display. |cffeda55fAlt-Click|r to issue a group reset. |cffeda55fShift-Click|r to reset."] = true,
	["Player"] = true,
	["#"] = true,
	["%s: %d"] = true,
} end)

L:RegisterTranslations("koKR", function() return {
	["Track"] = "측정",
	["What data types to track."] = "측정할 데이터를 선택합니다.",
	["Tooltip"] = "툴팁",
	["What data to show in the tooltip."] = "툴팁에 표시할 데이터를 선택합니다.",
	["Toggle whether to display data from %q in the tooltip."] = "툴팁에 %q 데이터 표시를 전환합니다.",
	["Colors"] = "색상",
	["Assign colors to the different data types."] = "각각의 데이터 종류에 색상을 지정합니다.",
	["What color to use for the %q data type."] = "%q 데이터에 사용할 색상을 지정합니다.",
	["Color the tooltip"] = "툴팁 색상",
	["Toggles whether or not to use these colors in the tooltip."] = "툴팁 색상 사용을 전환합니다.",

	["Bars"] = "바 설정",
	["Settings for the GUI."] = "GUI 설정",
	["Toggle the GUI."] = "GUI 전환",
	["Show"] = "보기",
	["Display"] = "화면",
	["What data to show in the GUI."] = "화면에 보일 데이터를 선택합니다.",
	["Texture"] = "텍스처",
	["Set the bar texture."] = "바 텍스처를 선택합니다.",
	["Scale"] = "크기",
	["Set the bar window scale."] = "창 크기를 설정합니다.",
	["Window Color"] = "창 색상",
	["Set the GUI Color"] = "GUI 색상을 설정합니다.",

	["Resetting data as requested by %s."] = "%s의 요청에 의해 데이터를 초기화합니다.",
	["Reset"] = "초기화",
	["Options for when to reset the data."] = "데이터 초기화에 대한 설정입니다.",
	["Reset now"] = "즉시 초기화",
	["Immediately resets your data."] = "당신의 데이터를 즉시 초기화합니다.",
	["Reset group"] = "파티 초기화",
	["Resets the Violation data for every group member. Only the group leader or an officer can do this."] = "모든 파티원에 대한 Violation 데이터를 초기화합니다. 단, 공대장 혹은 승급자 이어야 합니다.",
	["Leader resets"] = "리더 초기화",
	["Accept reset requests from raid leaders or officers."] = "공대장 혹은 오피서로 부터의 초기화 요청을 허용합니다.",
	["Auto-reset"] = "자동 초기화",
	["Entering combat"] = "전투 시작시",
	["Automatically reset the data when entering combat."] = "전투 시작시 자동으로 데이터를 초기화합니다.",
	["Leaving combat"] = "전투 종료시",
	["Automatically reset the data when leaving combat."] = "전투 종료시 자동으로 데이터를 초기화합니다.",
	["Print on reset"] = "초기화 시 출력",
	["Prints your current data locally when your data is reset."] = "당신의 데이터를 초기화할 때 현재 데이터를 출력합니다.",

	["Text"] = "텍스트",
	["What data to show on the FuBar panel, if available."] = "FuBar에 표시할 데이터를 선택합니다.",

	["Toggle whether to track %q."] = "%q 기록 전환",
	["Toggle whether to display data from %q on the FuBar panel."] = "FuBar에 %q 데이터 표시를 전환합니다.",

	["%s has been disabled, since it is not compatible with the latest module API."] = "%s|1이;가; 비활성화 되었습니다, 이후 해당 모듈은 최신 모듈 API와 호환되지 않습니다.",
	["No modules running, please enable one or more from the configuration menu."] = "실행중인 모듈이 없습니다. 환경설정 메뉴에서 하나 이상을 활성화하세요.",
	["No data gathered yet."] = "집계된 데이터가 없습니다.",
	["|cffeda55fClick|r to toggle the bar display. |cffeda55fAlt-Click|r to issue a group reset. |cffeda55fShift-Click|r to reset."] = "|cffeda55f클릭|r 바 표시 전환. |cffeda55fAlt-클릭|r 그룹 초기화. |cffeda55fShift-클릭|r 초기화.",
	["Player"] = "플레이어",
	["#"] = "#",
	["%s: %d"] = "%s: %d",
} end)

-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------
local tablet = AceLibrary("Tablet-2.0")
local BC = AceLibrary("Babble-Class-2.2")
local roster = AceLibrary("Roster-2.1")
local AceOO = AceLibrary("AceOO-2.0")
local dew = AceLibrary("Dewdrop-2.0")


local ChatThrottleLib = ChatThrottleLib
if type(ChatThrottleLib) ~= "table" or type(ChatThrottleLib.version) ~= "number" or ChatThrottleLib.version < 13 then
	error("Violation requires ChatThrottleLib 13 or higher.")
end

local SM = AceLibrary("SharedMedia-1.0")

local coloredNames = setmetatable({}, {__index =
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return key end
		self[key] = format("|cff%s%s|r", BC:GetHexColor(obj.class), key, "|r")
		return self[key]
	end
})

local RGB = setmetatable({}, {__index = 
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return nil end
		local r, g, b = BC:GetColor(obj.class)
		self[key] = {r, g, b}
		return self[key]
	end
})

local db = nil

local _G = getfenv(0)
local table_insert = table.insert
local string_gmatch = string.gmatch

local playerName = nil
local percentString = "(%.1f%%)"

local moduleVersion = 4

local runningModules = {}
local violationUsers = nil
local characterPets = {}
local moduleData = nil
local dataFormatter = {}
local characterDataSource = {}

local VIOLATION_LEVEL = 3
local SWS_LEVEL = 2
local DAMAGEMETERS_LEVEL = 1
local COMBATLOG_LEVEL = 0

local red = "Interface\\AddOns\\Violation\\icons\\red"
local blue = "Interface\\AddOns\\Violation\\icons\\blue"
local grey = "Interface\\AddOns\\Violation\\icons\\grey"

local printModuleList = {}
local whisperModuleList = {}
local whisperTargets = {
	player = {
		type = "text",
		name = "Player",
		desc = "Whisper the data to the specified player.",
		usage = "<playerName>",
		order = 1,
	},
	spacer = {
		type = "header",
		name = " ",
		order = 50,
	},
}

-----------------------------------------------------------------------
-- Local heap and utility functions
-----------------------------------------------------------------------

local new, del
do
	local cache = setmetatable({},{__mode="k"})
	function new(...)
		local t = next(cache)
		if t then
			cache[t] = nil
			for i = 1, select("#", ...) do
				table_insert(t, (select(i, ...)))
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == "table" then
			for k, v in pairs(t) do
				if type(v) == "table" then
					del(v)
				end
				t[k] = nil
			end
			cache[t] = true
		end
		return nil
	end
end

local function Colorize(text, r, g, b)
	return string.format("|cff%02x%02x%02x%s|r", r * 255, g * 255, b * 255, text)
end

-----------------------------------------------------------------------
-- Addon declaration
-----------------------------------------------------------------------
Violation = AceLibrary("AceAddon-2.0"):new(
	"AceEvent-2.0",
	"AceDB-2.0",
	"AceConsole-2.0",
	"AceComm-2.0",
	"FuBarPlugin-2.0",
	"AceModuleCore-2.0"
)
local Violation = Violation

Violation:SetModuleMixins("AceEvent-2.0")
Violation.hasIcon = true

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

local function get(key)
	return db[key]
end
local function set(key, val)
	db[key] = val
end

local options = {
	type = "group",
	handler = Violation,
	args = {
		track = {
			type = "group",
			name = L["Track"],
			desc = L["What data types to track."],
			disabled = "~IsActive",
			args = {}, -- Auto-populated by module toggles.
			order = 101,
			pass = true,
			get = function(mod)
				return db.enabledModules[mod]
			end,
			set = "ToggleModule",
		},
		fubar = {
			type = "group",
			name = L["Text"],
			desc = L["What data to show on the FuBar panel, if available."],
			disabled = function() return not FuBar end,
			args = {}, -- Auto-populated by module toggles.
			order = 102,
			pass = true,
			get = function(mod)
				return db.textModules[mod]
			end,
			set = function(mod, val)
				db.textModules[mod] = val
				Violation:UpdateText()
			end,
		},
		colors = {
			type = "group",
			name = L["Colors"],
			desc = L["Assign colors to the different data types."],
			pass = true,
			get = function(mod)
				if mod == "tooltip" then
					return db.colorTooltip
				else
					local tuple = db.colors[mod] or {r=1,g=1,b=1}
					return tuple.r or 1, tuple.g or 1, tuple.b or 1
				end
			end,
			set = function(mod, r, g, b)
				if mod == "tooltip" then
					db.colorTooltip = r
					Violation:UpdateTooltip()
				else
					local tuple = db.colors[mod]
					tuple.r = r
					tuple.g = g
					tuple.b = b
					Violation:ScheduleEvent("ColorUpdate", Violation.UpdateDisplay, 5, Violation)
					Violation:UpdateWindows(true)
				end
			end,
			args = {
				tooltip = {
					type = "toggle",
					name = L["Color the tooltip"],
					desc = L["Toggles whether or not to use these colors in the tooltip."],
					order = 1,
					disabled = true,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 50,
				},
			}, -- Auto-populated by module toggles.
			order = 103,
		},
		tooltip = {
			type = "group",
			name = L["Tooltip"],
			desc = L["What data to show in the tooltip."],
			args = {},
			order = 104,
			pass = true,
			get = function(mod)
				return db.tooltipModules[mod]
			end,
			set = function(mod, val)
				db.tooltipModules[mod] = val
				Violation:UpdateTooltip()
			end
		},
		newwindow = {
			type = "execute",
			name = "Create New Window",
			desc = "Create a new bar window",
			func = function()
				db.windows.count = db.windows.count + 1
				Violation:SetupWindows()
			end,
			order = 105
		},
		merge = {
			type = "toggle",
			name = "Merge pets",
			desc = "Merge pet data with master.",
			get = function() return db.mergePets end,
			set = function(value)
				db.mergePets = value
				Violation:UpdateDisplay()
				Violation:UpdateWindows()
			end,
			order = 110,
		},
		spacer = {
			type = "header",
			name = " ",
			order = 150,
		},
		resetPrint = {
			type = "toggle",
			name = L["Print on reset"],
			desc = L["Prints your current data locally when your data is reset."],
			get = get,
			set = set,
			passValue = "printOnReset",
			order = 200,
		},
		reset = {
			type = "group",
			name = L["Reset"],
			desc = L["Options for when to reset the data."],
			args = {
				now = {
					type = "execute",
					name = L["Reset now"],
					desc = L["Immediately resets your data."],
					func = "Reset",
					order = 100,
				},
				group = {
					type = "execute",
					name = L["Reset group"],
					desc = L["Resets the Violation data for every group member. Only the group leader or an officer can do this."],
					func = "ResetGroup",
					disabled = function()
						return not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader()
					end,
					order = 101,
				},
				leader = {
					type = "toggle",
					name = L["Leader resets"],
					desc = L["Accept reset requests from raid leaders or officers."],
					get = get,
					set = set,
					passValue = "leaderReset",
					order = 102,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 150,
				},
				autoheader = {
					type = "header",
					name = L["Auto-reset"],
					order = 200,
				},
				gaincombat = {
					type = "toggle",
					name = L["Entering combat"],
					desc = L["Automatically reset the data when entering combat."],
					get = function() return Violation:IsEventRegistered("PLAYER_REGEN_DISABLED") end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_DISABLED")
						end
						db.gainCombat = v
					end,
					order = 201,
				},
				leavecombat = {
					type = "toggle",
					name = L["Leaving combat"],
					desc = L["Automatically reset the data when leaving combat."],
					get = function() return Violation:IsEventRegistered("PLAYER_REGEN_ENABLED") end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_ENABLED")
						end
						db.leaveCombat = v
					end,
					order = 202,
				},
			},
			order = 201,
		},
		report = {
			type = "group",
			name = "Report",
			desc = "Report your current data to a channel.",
			args = {
				player = {
					type = "group",
					name = "Whisper",
					desc = "Whisper the data to someone.",
					args = whisperModuleList,
					order = 1,
				},
				say = {
					type = "group",
					name = "Say",
					desc = "Print module data to say.",
					pass = true,
					func = function(modName) Violation:PrintModuleDataToChannel(modName, "SAY") end,
					args = printModuleList,
					order = 2,
				},
				guild = {
					type = "group",
					name = "Guild",
					desc = "Print module data on guild channel.",
					pass = true,
					func = function(modName) Violation:PrintModuleDataToChannel(modName, "GUILD") end,
					args = printModuleList,
					order = 2,
				},
				officer = {
					type = "group",
					name = "Officer",
					desc = "Print module data on officer channel.",
					pass = true,
					func = function(modName) Violation:PrintModuleDataToChannel(modName, "OFFICER") end,
					args = printModuleList,
					order = 2,
				},
				party = {
					type = "group",
					name = "Party",
					desc = "Print module data on guild channel.",
					pass = true,
					func = function(modName) Violation:PrintModuleDataToChannel(modName, "PARTY") end,
					args = printModuleList,
					order = 2,
				},
				raid = {
					type = "group",
					name = "Raid",
					desc = "Print module data on raid channel.",
					pass = true,
					func = function(modName) Violation:PrintModuleDataToChannel(modName, "RAID") end,
					args = printModuleList,
					order = 2,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 50,
				},
				limit = {
					type = "range",
					name = "Lines",
					desc = "Maximum number of lines to print.",
					max = 40,
					min = 1,
					step = 1,
					get = get,
					set = set,
					passValue = "maxPrintLines",
					order = 100,
				},
			},
			order = 202,
		},
		spying = {
			type = "group",
			name = "Spying",
			desc = "Spying allows Violation to read data sent by other meter addons. " 
			       .. Colorize("Although it receives their data, it will not send anything back to them.",1,0.5,0),
			order = 203,
			args = {
				dmm = {
					type = "toggle",
					name = "DamageMeters",
					desc = "Spy DamageMeters.",
					get = function() return db.spying.damageMeters end,
					set = function(v)
						db.spying.damageMeters = v 
						Violation:UpdateSpying()
					end,
				},
				sws = {
					type = "toggle",
					name = "SWStats",
					desc = "Spy SWStats.",
					get = function() return db.spying.swStats end,
					set = function(v)
						db.spying.swStats = v 
						Violation:UpdateSpying()
					end,
				},
			},
		},
		syncInfo = {
			type = "execute",
			name = "Syncing report",
			desc = "Display which player are actually sending data.",
			func = "ReportUsers",
			order = 204,
		},
	},
}

-- options for the rightclick bars menu
local baroptions = options.args.bars

-----------------------------------------------------------------------
-- Module prototype
-----------------------------------------------------------------------

Violation.modulePrototype.version = -1

local incompatible = {}
local registered = {}
function Violation.modulePrototype:OnInitialize()
	if type(self.name) ~= "string" then
		error("A module is missing its name field.")
	elseif type(self.version) ~= "number" then
		error(self.name .. " doesn't have a version field.")
	elseif registered[self.name] then
		error("A module has been initialized twice for some reason.")
	end

	if self.version == moduleVersion then
		Violation:OnRegister(self)
		registered[self.name] = true
	else
		incompatible[self] = true
		Violation:ModuleIncompatible(self)
	end
end

function Violation.modulePrototype:OnEnable(first)
	if incompatible[self] or not Violation.db.char.enabledModules[self.name] then return end

	if not runningModules[self.name] then
		-- Always reset our data when we enable.
		self:OnReset()

		if type(self.OnModuleEnable) == "function" then
			self:OnModuleEnable()
		end

		runningModules[self.name] = self
		-- options.args.bars.args.display.validate[self.name] = self:GetDisplayName()

		-- Violation:SendCommMessage("GROUP", "ModuleEnabled", self.name)
		Violation:TriggerEvent("Violation_ModuleEnabled", self.name)
	end
end

function Violation.modulePrototype:OnDisable()
	if runningModules[self.name] then
		if type(self.OnModuleDisable) == "function" then
			self:OnModuleDisable()
		end

		runningModules[self.name] = nil
		options.args.bars.args.display.validate[self.name] = nil

		local availableMod
		for name in pairs(runningModules) do
			availableMod = name
			break
		end

		for k, w in ipairs( db.windows ) do
			if w.module == self.name then
				if availableMod then
					w.module = availableMod
				else
					Violation:DeleteWindow( self.windows[k] )
				end
			end
		end
		
		Violation:UpdateWindows()
		
		-- Violation:SendCommMessage("GROUP", "ModuleDisabled", self.name)
		Violation:TriggerEvent("Violation_ModuleDisabled", self.name)
	end
end

function Violation.modulePrototype:GetMergedData(...)
	local sum = 0
	for i = 1,select("#", ...) do
		sum = sum + (self:GetData(select(i, ...)) or 0)
	end
	return sum
end

function Violation.modulePrototype:IsShown()
	local name = self.name
	if runningModules[name] then
		if db.textModules[name] or db.tooltipModules[name] then
			return true
		end
		for k, w in ipairs( Violation.windows ) do
			if w.options.module == name then
				return true
			end
		end
	end
	return false
end

function Violation.modulePrototype:RequestDisplayRefresh()
	if self:IsShown() then
		self:TriggerEvent('Violation_UpdateDisplay', self.name)
	end
end

function Violation.modulePrototype:new(...) return new(...) end
function Violation.modulePrototype:del(t) return del(t) end

-----------------------------------------------------------------------
-- Implicit module interface
-----------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local IViolation = AceOO.Interface{

	-- Optional methods:
	--
	-- -- Called when the module is enabled by the core. Any event registration
	-- -- should happen here.
	-- OnModuleEnable
	-- OnModuleDisable
	--
	-- -- Called when the module is initially registered.
	-- OnModuleRegister

	-- Must return the current module revision as a number.
	GetRevision = "function",

	-- Must return the module name as it should be displayed in the menu, for
	-- example "Damage done".
	GetDisplayName = "function",

	-- Must return the module name as it should be displayed in a condensed
	-- display, like "PlayerDPS" should return "DPS".
	GetShortDisplayName = "function",

	-- Must return whether the modules broadcastdata is dirty or not. If it is, a
	-- broadcast will be scheduled by the core.
	IsBroadcastDirty = "function",

	-- Must return wether the modules data is dirty or not. If it is a
	-- diplay update will be scheduled by the core.
	IsDirty = "function",

	-- This function is must reset the modules dirty flag that is returned by IsDirty.
	ResetDirty = "function",

	-- Must return the data that should be broadcasted to the other players.
	-- Must return 1 number only. This function should reset the broadcastdirty flag as well.
	GetBroadcastData = "function",

	-- Invoked when remote data is received for this module. Usually invoked by
	-- the core, but can also be invoked by other modules when they parse data
	-- from other addons, for example.
	-- Will only get two arguments, the first will be the person the data
	-- belongs to and the second will be a number.
	OnRemoteData = "function",

	-- Invoked by the core when all data should be reset.
	OnReset = "function",

	-- Must return a universal unique ID that will be used for this module when
	-- sending data (where the universe is Violation)
	-- (Not used now, we use a memoization of the module name instead.
	-- GetUUID = "function",

	-- Must return the current dataset for this module. If an argument is given,
	-- it can be assumed to be a player name, and only the data (a number) for
	-- that player (or nil) must be returned.
	GetData = "function",
}

-----------------------------------------------------------------------
-- Addon initialization
-----------------------------------------------------------------------

function Violation:OnInitialize()
	local revision = tonumber(("$Revision: 38980 $"):sub(12, -3)) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.VIOLATION_VERSION = self.version
	_G.VIOLATION_REVISION = self.revision

	self:RegisterDB("ViolationDB", "ViolationPerCharDB", "char")
	self:RegisterDefaults("char", {
		enabledModules = {
			["*"] = true,
		},
		textModules = {
			DPS = true,
		},
		tooltipModules = {
			DPS = true,
		},
		colors = {
			["**"]            = { r = 1  , g = 1  , b = 1   },
			
			["Damage"]        = { r = 0.8, g = 0.3, b = 0.3 },
			["DamageTaken"]   = { r = 0.8, g = 0.5, b = 0.3 },
			["DPS"]           = { r = 0.8, g = 0.8, b = 0.3 },
			["Healing"]       = { r = 0.5, g = 0.8, b = 0.3 },
			["HealingTaken"]  = { r = 0.3, g = 0.8, b = 0.3 },
			["HPS"]           = { r = 0.3, g = 0.8, b = 0.5 },
			["Overhealing"]   = { r = 0.3, g = 0.8, b = 0.8 },
			["Threat"]        = { r = 0.3, g = 0.5, b = 0.8 },
			["Target Threat"] = { r = 0.3, g = 0.3, b = 0.8 },
			["Decurse"]       = { r = 0.5, g = 0.3, b = 0.8 },
			["HitCritRate"]   = { r = 0.8, g = 0.3, b = 0.8 },
			["Banzai"]        = { r = 0.8, g = 0.3, b = 0.5 },
			
		},
		colorTooltip = true,
		gainCombat = nil,
		leaveCombat = nil,
		leaderReset = true,
		printOnReset = true,
		mergePets = true,
		
		-- Spying
		spying = {
			damageMeters = true,
			swStats = true,
		},

		windows = {
			count = 1,
			["**"] = {
				width = 200,
				height = 200,
				nrBars = 10,
				scale = 1,
				shown = true,
				module = "Damage",
				locked = false,
				useModuleColor = true,
				color = { r = 1, g = 0, b = 0 },
				barTexture = "Blizzard",
			},
		},
		
		-- Bar options
		barModule = "Damage",
		nrBars = 10,
		barTotalHeight = 200,
		barTotalWidth = 154,
		barTexture = "Blizzard",
		barScale = 1,
		showBars = true,
		lockBars = nil,
		useModuleColor = true,
		windowColor = {
			r = 1, g = 0, b = 0,
		},
	})

	db = self.db.char

	self.OnMenuRequest = options

	self:RegisterChatCommand("/violation", self.OnMenuRequest, "VIOLATION")

	self:SetCommPrefix("Violation")
	self:SetDefaultCommPriority("BULK")

	self:SetIcon(grey)

	playerName = UnitName("player")
	
	-- small change for the window gui
	self.runningModules = runningModules
end

function Violation:OnEnable(first)
	if first then
		local memoizations = new()
		for name in self:IterateModules() do
			table_insert(memoizations, name)
		end
		table_insert(memoizations, "LeaderReset")
		table_insert(memoizations, "ViolationPing")
		table_insert(memoizations, "ViolationPong")
		self:RegisterMemoizations(memoizations)
		memoizations = del(memoizations)
	end

	-- Enable/disable all modules.
	for name, mod in self:IterateModules() do
		self:ToggleModuleActive(name, db.enabledModules[name])
	end

	self:SetupWindows()
	
	self:Reset()

	self:RegisterComm("Violation", "GROUP", "OnComm")
	self:RegisterComm("Violation", "WHISPER", "OnPrivateComm")

	self:ScheduleRepeatingEvent("Broadcast", self.Broadcast, 5, self)
	self:ScheduleRepeatingEvent("DisplayUpdate", self.UpdateDisplayIfDirty, 1, self)	-- 2 is too long for threat, at least early in combat /Mikk

	if db.gainCombat then
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
	end

	if db.leaveCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
	end

	self:RegisterBucketEvent("RosterLib_RosterUpdated", 2, function()
		self:SendCommMessage("GROUP", "ViolationPing")
		self:UpdatePartyMenu()
		self:OnRosterUpdate()
	end)

	self:UpdateSpying()

	self:RegisterBucketEvent("Violation_UpdateDisplay", 0.1, "UpdateDisplayIfDirty")
	
	self:SetIcon(red)
	self:OnRosterUpdate()
end

function Violation:OnDisable()
	-- Disable all modules.
	for name, mod in pairs(runningModules) do
		self:ToggleModuleActive(name, false)
	end

	self:SetIcon(grey)
end

function Violation:Reset()
	for name, mod in pairs(runningModules) do
		if db.printOnReset then
			local data = mod:GetData(playerName)
			if type(data) == "number" and data > 0 then
				local name = mod:GetDisplayName()
				local color = db.colors[name]
				self:Print(Colorize(string.format(L["%s: %d"], name, data), color.r, color.g, color.b))
			end
		end
		mod:OnReset()
	end
	self:UpdateDisplay()
	self:UpdateWindows()
end

function Violation:new(...) return new(...) end
function Violation:del(t) return del(t) end
Violation.RGB = RGB

-----------------------------------------------------------------------
-- Module handling
-----------------------------------------------------------------------

function Violation:OnRegister(mod)
	if not AceOO.inherits(mod, IViolation) then
		error(string.format("The module %q does not inherit from the IViolation interface.", tostring(mod.name)))
	end

	local moduleName = mod.name
	local displayName = mod:GetDisplayName()

	local function disabled()
		return not runningModules[moduleName]
	end

	options.args.track.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to track %q."], displayName),
	}

	options.args.colors.args[moduleName] = {
		type = "color",
		name = displayName,
		desc = string.format(L["What color to use for the %q data type."], displayName),
		order = 100,
		hidden = disabled,
	}

	options.args.fubar.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to display data from %q on the FuBar panel."], displayName),
		hidden = disabled,
	}

	options.args.tooltip.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to display data from %q in the tooltip."], displayName),
		hidden = disabled,
	}

	printModuleList[moduleName] = {
		name = displayName,
		desc = format("Send %s data.", displayName),
		type = "execute",
		hidden = disabled,
	}

	whisperModuleList[moduleName] = {
		type = "group",
		name = displayName,
		desc = format("Send %s data to player.", displayName),
		pass = true,
		get = false,
		set = function(passValue, playerName) return self:PrintModuleDataToChannel(moduleName, "WHISPER", playerName) end,
		func = function(playerName) return self:PrintModuleDataToChannel(moduleName, "WHISPER", playerName) end,
		args = whisperTargets,
		hidden = disabled,
	}

	if mod.dataFormatter then
		if type(mod.dataFormatter) == "function" then
			dataFormatter[mod.name] = mod.dataFormatter
		elseif type(mod.dataFormatter) == "string" then
			dataFormatter[mod.name] = function(v) return mod.dataFormatter:format(v) end
		end
	end

	if type(mod.OnModuleRegister) == "function" then
		mod:OnModuleRegister()
	end
end

function Violation:GetModuleVersion()
	return moduleVersion
end

function Violation:ModuleIncompatible(mod)
	self:ToggleModuleActive(mod, false)
	self:Print(string.format(L["%s has been disabled, since it is not compatible with the latest module API."], mod.name))
end

function Violation:ToggleModule(mod, val)
	db.enabledModules[mod] = val
	Violation:ToggleModuleActive(mod, val)
	Violation:UpdateDisplay()
end

-----------------------------------------------------------------------
-- Pet merging stuff
-----------------------------------------------------------------------

function Violation:OnRosterUpdate()
	-- Update characterPets for named pets (warlocks' and hunters' ones)

	local seen = new()
	local dirty = false

	-- Scan roster
	for unit in roster:IterateRoster() do
		-- Gotya
		seen[unit.name] = true
		characterDataSource[unit.name] = unit.name

		-- Record hunter and warlock pets
		if unit.class == "WARLOCK" or unit.class == "HUNTER" then
			-- Check for current pet
			local pet = roster:GetUnitObjectFromUnit(roster:GetPetFromOwner(unit.unitid))
			if pet then
				if not characterPets[unit.name] then
					characterPets[unit.name] = new()
				end
				if not characterPets[unit.name][pet.name] then
					-- This is very ugly, but needed...
					characterPets[unit.name][pet.name] = true
					table_insert(characterPets[unit.name], pet.name)
					dirty = true
				end
			end
		-- Ignore pets of other classes
		elseif characterPets[name] then
			characterPets[name] = del(characterPets[name])
			dirty = true
		end
	end

	-- Remove old data
	for name in pairs(characterPets) do
		if not seen[name] then
			characterPets[name] = del(characterPets[name])
			characterDataSource[name] = nil
			dirty = true
		end
	end
	
	-- Link pets to master
	for master, pets in pairs(characterPets) do
		for i,pet in ipairs(pets) do
			characterDataSource[pet] = master
		end
	end

	seen = del(seen)

	if dirty then
		self:UpdateDisplay()
		self:UpdateWindows()
	end
end

function Violation:GetModuleData(modName, charName)
	local mod = runningModules[modName]
	if not mod then
		return
	end
	if db.mergePets and mod.canMergeData then
		-- Specific player data
		if charName then
			-- Merge if necessary
			if characterPets[charName] then
				return mod:GetMergedData(charName, unpack(characterPets[charName]))
			end
		-- All module data
		else
			local seen = new()
			-- Clean previous data table
			moduleData = del(moduleData)
			moduleData = new()

			-- Clone module data
			for k,v in pairs(mod:GetData()) do
				moduleData[k] = v
			end

			-- Merge data where needed
			for main, others in pairs(characterPets) do
				-- Delete pets from moduleData
				for k, v in pairs( others ) do
					moduleData[v] = nil
				end
				moduleData[main] = mod:GetMergedData(main, unpack(others))
			end

			return moduleData
		end

	end
	-- Fallback to old, plain data
	return mod:GetData(charName)
end

-----------------------------------------------------------------------
-- 
-----------------------------------------------------------------------

local function Sorter(a, b) return a[2] > b[2] end

function Violation:PublishModuleData(modName, rowCallback, count, offset, window)
	local mod = runningModules[modName]
	if not mod then
		return
	end
	local data = self:GetModuleData(modName)
	assert(type(data) == "table", "Some module provided us with nil data.")

	local dataInfo = new()
	local tbl = new()
	dataInfo.total = 0
	dataInfo.maxValue = 0
	for k, v in pairs(data) do
		if type(v) == "number" and v > 0 then
			dataInfo.total = dataInfo.total + v
			dataInfo.maxValue = math.max(dataInfo.maxValue, v)
			table_insert(tbl, new(k, v))
		end
	end

	local tblCount = #tbl

	count = math.min(count or tblCount, tblCount)
	offset = math.max(math.min(offset or 0, tblCount-count), 0)

	local n = 0
	if tblCount > 0 then
		local percentage = nil

		if tblCount > 1 then
			table.sort(tbl, Sorter)

			if not mod.hidePercentage then
				percentage = new()
				for i, v in ipairs(tbl) do
					table_insert(percentage, (v[2] * 100) / dataInfo.total)
				end
			end
		end

		dataInfo.hasMergedData = db.mergePets and mod.canMergeData and next(characterPets) and true
		dataInfo.rowCount = tblCount
		dataInfo.module = mod
		dataInfo.formatter = dataFormatter[modName] or tostring
		dataInfo.hasPercentage = percentage

		for i = 1, count do
			local rank = offset + i
			local name, value = unpack(tbl[rank])
			if not name then
				break
			end
			--[[
			local source = characterDataSource[name]
			if source and not Violation:UserViolation(source, modName, VIOLATION_LEVEL) then
				name = name .. '*'
			end
			--]]
			if window then
				rowCallback(window, i, rank, name, value, percentage and percentage[rank], dataInfo)
			else
				rowCallback(i, rank, name, value, percentage and percentage[rank], dataInfo)
			end
			n = n + 1
		end

		percentage = del(percentage)
	end

	tbl = del(tbl)

	-- We don't want to del tables contained in dataInfo
	-- We really don't want to...
	for k,v in pairs(dataInfo) do
		if type(v) == "table" then
			dataInfo[k] = nil
		end
	end

	dataInfo = del(dataInfo)

	return n, offset
end

-----------------------------------------------------------------------
-- Channel output
-----------------------------------------------------------------------

function Violation:PrintModuleDataToChannel(modName, channel, target)

	-- This should stay here (depends on channel)
	local function doSend(msg, ...)
		if select("#", ...) > 0 then
			msg = msg:format(...)
		end
		ChatThrottleLib:SendChatMessage("BULK", nil, msg, channel, nil, target)
	end

	self:PublishModuleData(
		modName,
		function(i, rank, name, value, percentage, dataInfo)
			if not dataInfo.header then
				local name = dataInfo.module:GetDisplayName()
				if dataInfo.hasMergedData then
					doSend("==== %s (pets merged) ====", name)
				else
					doSend("==== %s ====", name)
				end
				dataInfo.header = true
			end
			local value = dataInfo.formatter(value)
			if percentage then
				doSend("%d. %s - %s "..percentString, rank, name, value, percentage)
			else
				doSend("%d. %s - %s", rank, name, value)
			end
		end,
		db.maxPrintLines
	)
end

-----------------------------------------------------------------------
-- UI
-----------------------------------------------------------------------

local classOrders = {
	WARRIOR = 100,
	ROGUE = 200,
	SHAMAN = 300,
	HUNTER = 400,
	MAGE = 500,
	WARLOCK = 600,
	DRUID = 700,
	PRIEST = 800,
	PALADIN = 900,
}

function Violation:UpdatePartyMenu()
	local seen = new()

	for unit in roster:IterateRoster() do
		if unit and unit.class and unit.class ~= "PET" then
			local name = unit.name
			seen[name] = true
			if not whisperTargets[name] then
				local option = new()
				option.type = "execute"
				option.name = coloredNames[name]
				option.desc = string.format("Send module data to %s.", name)
				option.passValue = name
				option.order = classOrders[unit.class]
				whisperTargets[name] = option
			end
		end
	end

	for k,o in pairs(whisperTargets) do
		if o.passValue and not seen[o.passValue] then
			whisperTargets[k] = del(whisperTargets[k])
		end
	end

	seen = del(seen)
end

function Violation:OnClick()
	if IsAltKeyDown() then
		self:ResetGroup()
	elseif IsShiftKeyDown() then
		self:Reset()
--	else
--		self:ToggleBarDisplay(not db.showBars)		
	end
end

function Violation:UpdateDisplayIfDirty()
	local dirty = nil
	local dirtpile = new()
	for k, v in pairs(db.tooltipModules) do
		if v and runningModules[k] and runningModules[k]:IsDirty() then
			table_insert( dirtpile, k )
			dirty = true
		end
	end
	if dirty then
		self:UpdateDisplay()
	end
	
	self:UpdateDirtyWindows()

	for k, v in ipairs( dirtpile ) do
		runningModules[v]:ResetDirty()
	end
	dirtpile = del(dirtpile)
end

local tooltipColors = {
	"child_textR", 1,
	"child_textG", 1,
	"child_textB", 0,
	"child_text2R", 1,
	"child_text2G", 1,
	"child_text2B", 1,
	"child_text3R", 1,
	"child_text3G", 1,
	"child_text3B", 1,
	"child_text4R", 0.65,
	"child_text4G", 0.65,
	"child_text4B", 0.65
}

-- row adding callback
local function rowCallback(i, rank, name, value, percentage, dataInfo)
	-- Create header on the fly
	if not dataInfo.cat then
		local displayName = dataInfo.module:GetDisplayName()
		local tblCount = dataInfo.rowCount
		dataInfo.cat = tablet:AddCategory(
			"columns", 2 + (tblCount > 1 and 1 or 0) + (dataInfo.hasPercentage and 1 or 0),
			"text", tblCount > 1 and L["#"] or displayName,
			"text2", tblCount > 1 and L["Player"] or "",
			"text3", displayName,
			"child_justify", "LEFT",
			"child_justify2", tblCount > 1 and "LEFT" or "RIGHT",
			"child_justify3", "RIGHT",
			"child_justify4", "RIGHT",
			unpack(tooltipColors)
		)
	end

	-- Add row
	if dataInfo.rowCount > 1 then
		dataInfo.cat:AddLine(
			"text", tostring(rank) .. ".",
			"text2", coloredNames[name],
			"text3", dataInfo.formatter(value or 0),
			"text4", percentage and percentString:format(percentage)
		)
	else
		dataInfo.cat:AddLine(
			"text", coloredNames[name],
			"text2", dataInfo.formatter(value or 0)
		)
	end
end

function Violation:OnTooltipUpdate()
	local no = true
	for k, v in pairs(db.tooltipModules) do
		if v and runningModules[k] then
			no = nil
			self:PublishModuleData(k, rowCallback, 10, 0)
		end
	end
	if no then
		tablet:AddCategory("columns", 1):AddLine("text", L["No modules running, please enable one or more from the configuration menu."])
	end
	tablet:SetHint(L["|cffeda55fClick|r to toggle the bar display. |cffeda55fAlt-Click|r to issue a group reset. |cffeda55fShift-Click|r to reset."])
end

function Violation:OnTextUpdate()
	local hasText = nil
	for k, v in pairs(db.textModules) do
		if v and runningModules[k] then
			local data = self:GetModuleData(k, playerName)
			if type(data) == "number" and data > 0 then
				local shortName = runningModules[k]:GetShortDisplayName()
				local color = db.colors[k]
				data = (dataFormatter[k] or tostring)(data)
				local text = Colorize(string.format("%s: %s", shortName, data), color.r, color.g, color.b)
				if not hasText then hasText = new(text)
				else table_insert(hasText, text) end
			end
		end
	end

	if not hasText then
		self:SetText(self.name)
	else
		self:SetText(table.concat(hasText, " "))
		hasText = del(hasText)
	end
end

-----------------------------------------------------------------------
-- Data broadcasting and recieving
-----------------------------------------------------------------------

function Violation:Broadcast()
	local tbl, data = nil, nil
	for name, mod in pairs(runningModules) do
		if mod:IsBroadcastDirty() then
			if not tbl then tbl = new() end
			table_insert(tbl, name)
			table_insert(tbl, mod:GetBroadcastData())
		end
	end
	if tbl then
		self:SendCommMessage("GROUP", "B", unpack(tbl))
	end
	tbl = del(tbl)
end

function Violation:ResetGroup()
	if not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader() then
		self:Print("Only the group leader or an officer can do that.")
		return
	end

	self:SendCommMessage("GROUP", "LeaderReset")
	self:Reset()
end

function Violation:OnPrivateComm(prefix, sender, distribution, command, ...)
	if distribution ~= "WHISPER" or type(command) ~= "string" or type(sender) ~= "string" then return end
	if command == "ViolationPong" then
		if type(violationUsers) ~= "table" then
			violationUsers = new()
		end
		if violationUsers[sender] then
			violationUsers[sender] = del(violationUsers[sender])
		end
		violationUsers[sender] = new()
		violationUsers[sender].version = select(1,...)
		violationUsers[sender].revision = select(2,...)
		for i = 3, select("#", ...) do
			local x = (select(i, ...))
			if x then
					violationUsers[sender][x] = VIOLATION_LEVEL
			end
		end
	end
end

function Violation:OnComm(prefix, sender, distribution, command, ...)
	if distribution ~= "GROUP" or type(command) ~= "string" then return end

	if command == "B" then
		for i = 1, select("#", ...), 2 do
			local name, value = select(i, ...)
			if value and tonumber(value) > 0 then
				self:SetUserViolation(sender, name, VIOLATION_LEVEL)
				if runningModules[name] then
					runningModules[name]:OnRemoteData(sender, value)
				end
			end
		end
	elseif command == "LeaderReset" and db.leaderReset then
		local obj = roster:GetUnitObjectFromName(sender)
		if obj.rank > 0 then
			self:Print(string.format(L["Resetting data as requested by %s."], sender))
			self:Reset(sender)
		end
	elseif command == "ViolationPing" then
		local tbl = new()
		for k in pairs(runningModules) do
			table_insert(tbl, k)
		end
		self:SendCommMessage("WHISPER", sender, "ViolationPong", self.version, self.revision, unpack(tbl))
		tbl = del(tbl)
	end
end

function Violation:UserViolation(user, mod, level)
	if type(violationUsers) ~= "table" or type(violationUsers[user]) ~= "table" then return end
	if not level then level = COMBATLOG_LEVEL end
	if mod and violationUsers[user][mod] and violationUsers[user][mod] > level then
		return true
	elseif not mod and violationUsers[user] then
		return true
	end
	return false
end

function Violation:SetUserViolation(user, mod, level)
	if type(violationUsers) ~= "table" then
		violationUsers = new()
	end
	if type(violationUsers[user]) ~= "table" then
		violationUsers[user] = new()
	end
	if level > (violationUsers[user].globalLevel or 0) then
		violationUsers[user].globalLevel = level
	end
	violationUsers[user][mod] = level
end

-- Check if :
-- * both target and source are in party,
-- * source may provide data for target,
-- * source has sufficient level to provide data for target, according to module and level
-- source defaults to local player,
-- level defaults to VIOLATION_LEVEL for local player and to COMBATLOG_LEVEL for others
function Violation:AcceptDataFromSource(destination, module, source, level)
	if destination == ParserLib_SELF then destination = playerName end

	-- We're only interested of destination inside the party
	local regularSource = characterDataSource[destination]
	if not regularSource then
		return false
	end

	-- Source is local player
	if not source or source == ParserLib_SELF or source == playerName then
		
		-- Player is the regular source of target: always accept
		if regularSource == playerName then
			return true
		end
		-- Player isn't the regular source, do the check anyway

	-- Source isn't the regular source and isn't local player, drop data			
	elseif regularSource ~= source then
		return false
	end

	-- Check if regular source has the sufficient level	
	return not self:UserViolation(regularSource, module, level)
end

function Violation:ReportUsers()
	if violationUsers and next(violationUsers) then
		for unit in roster:IterateRoster() do
			local user = unit.name
			if user ~= playerName then
				local data = violationUsers and violationUsers[user]
				local level = data and data.globalLevel
				local msg = user..": "
				if level == VIOLATION_LEVEL then
					if data.version then
						msg = msg .. "Violation v%s".. data.version
					end
					local tbl = new()
					for mod,v in pairs(data) do
						if type(mod) == "string" and v == VIOLATION_LEVEL and mod ~= 'revision' and mod ~= 'version' then
							table_insert(tbl, mod)
						end
					end
					if #tbl then
						msg = string.format("%s, modules: %s", msg, table.concat(tbl, ", "))
					end
					tbl = del(tbl)
				elseif level == SWS_LEVEL then
					msg = msg .. "spying SWStats"
				elseif level == DAMAGEMETERS_LEVEL then
					msg = msg .. "spying DamageMeters"
				else
					msg = msg .. "combat log parsing"
				end
				self:Print(msg)
			end
		end
	else
		self:Print("No syncing information")
	end
end

-----------------------------------------------------------------------
-- Spying support
-----------------------------------------------------------------------

function Violation:UpdateSpying() 
	
	-- (Un)register events
	if (db.spying.damageMeters or db.spying.swStats) then
		if not self:IsEventRegistered("CHAT_MSG_ADDON") then
			self:RegisterEvent("CHAT_MSG_ADDON", "OnRawComm")
		end
	elseif self:IsEventRegistered("CHAT_MSG_ADDON") then
		self:UnregisterEvent("CHAT_MSG_ADDON")
	end
	
	-- Reset UserViolation data if needed
	if violationUsers then		
		for user,data in pairs(violationUsers) do
			local maxLevel = COMBATLOG_LEVEL
			for mod,level in pairs(data) do
				if mod ~= "version" and mod ~= "revision" and type(level) == "number" then
					if (level == SWS_LEVEL and not db.spying.swStats)
					   or (level == DAMAGEMETERS_LEVEL and not db.spying.damageMeters) then
						violationUsers[user][mod] = nil
					elseif level > maxLevel then
						maxLevel = level
					end
				end
			end
			if maxLevel == COMBATLOG_LEVEL then
				violationUsers[user] = del(violationUsers[user])
			else
				data.globalLevel = maxLevel
			end
		end
	end
end

Violation.commSpies = {}

function Violation:OnRawComm(prefix, message, distribution, sender)
	local spy = self.commSpies[prefix]

	if not spy or (distribution ~= "PARTY" and distribution ~= "RAID" and distribution ~= "BATTLEGROUND") then
		return
	end
	
	if sender == playerName or not roster:GetUnitObjectFromName(sender) then
		return
	end
	return spy(self, message, distribution, sender)
end

-----------------------------------------------------------------------
-- DamageMeters supports
-----------------------------------------------------------------------

-- Damage meter sends a triplet (total/hit/crit) for each of these data
local damageMeterInfo = {
	-- Protocol version 12
	[12] = { "Damage", "Healing", "DamageTaken", "HealingTaken", "Decurse", "Overhealing" },
}

-- Hold current DM session values
local damageMeterSession

function Violation.commSpies:DamageMeters(message, distribution, sender)
	if not db.spying.damageMeters then
		return
	end

	-- Parse message
	local command, version, arguments = select(3, message:find("(%w+)_(%d+)_ (.*)$") )
	
	-- Check version
	version = version and tonumber(version)
	if not version or not damageMeterInfo[version] then
		return
	end
	
	if command == "REQ_CLEAR" then
		-- Clear request
		local sessionId = select(3, arguments:find("^(.+ %d+)$") )
		if sessionId then
			damageMeterSession = del(damageMeterSession)
			damageMeterSession = new()
			damageMeterSession.sessionId = sessionId
		end
		
	elseif command == "REQ_SYNC" then		
		-- Sync request, check session
		
		local sessionId = select(3, arguments:find("^<(.+ %d+)>$") )
		if sessionId and damageMeterSession and damageMeterSession.sessionId ~= sessionId then
			-- Sessions mismatch, drop all data
			damageMeterSession = del(damageMeterSession)
		end
		
	elseif command == "SYNC" then
		-- Sync data
		local name, data = select(3, arguments:find("^(%S+) (.+)$") )
	
		-- Data parsing
		local dmi = damageMeterInfo[version]
		local index = 1
		for value in data:gmatch("(%d+) %d+ %d+") do
			if index > #dmi then
				break
			end
			value = tonumber(value)
			if value > 0 then
				local modName = dmi[index]

				if self:AcceptDataFromSource(name, modName, sender, DAMAGEMETERS_LEVEL) then
					self:SetUserViolation(sender, modName, DAMAGEMETERS_LEVEL)
					
					-- Vivify session data
					if not damageMeterSession then
						damageMeterSession = new()
					end
					if not damageMeterSession[modName] then
						damageMeterSession[modName] = new()
					end
				
					-- Only feed Violation if we are in a session
					if damageMeterSession.sessionId and runningModules[modName] then
						local oldValue = (damageMeterSession[modName][name] or 0)
						runningModules[modName]:OnRemoteData(name, value-oldValue)
					end
				
					-- Record value for future use
					damageMeterSession[modName][name] = value
				end
				
			end
			index = index + 1
		end
		
		-- Create a fake session if needed (until someone send us an real session id)
		if damageMeterSession and not damageMeterSession.sessionId then
			damageMeterSession.sessionId = "FAKE-"..tostring(damageMeterSession)
		end
	end
end


-- Hold current SWS session values
local SWSSession = {}
local SWSSyncMessage = "(%d+)"..string.char(31).."(.+)"
local SWSSplitMessage = "(.+)"..string.char(30).."(.+)"..string.char(30).."(.+)"
local SWSSplitSync = "[^"..string.char(29).."]+"

local SWSmodulesOUT = {
	"Damage",
	"Healing",
	false, -- "Deaths",
	"Decurse",
	false, --"TotalDmgTicks",
	"OverHealing",
	false, -- "HealInFight",
	false, -- "OverHealInFight",
	false, -- "DOTTicks",
	false, -- "DOTValue", -- SWS records dots separate from damage, we handle this in the main sync handler
	false, -- "TotalFightTime", 
}

local SWSmodulesIN = {
	"DamageTaken",
	"HealingTaken",
	false, -- "Deaths",
	false, -- "Decurse",
	false, --"TotalDmgTicks",
	false, -- "OverHealing",
	false, -- "HealInFight",
	false, -- "OverHealInFight",
	false, -- "DOTTicks",
	false, -- "DOTValue", -- SWS records dots separate from damage, we handle this in the main sync handler
	false, -- "TotalFightTime", 
}

-- If the module is currently running has already been asserted when calling this function
local function AddSWSData( sender, name, modName, value )
	if Violation:AcceptDataFromSource(name, modName, sender, SWS_LEVEL) then
		Violation:SetUserViolation(sender, modName, SWS_LEVEL)
		-- self:Print( "Received "..tostring(value).." SWS from "..sender.." for "..modName )
		if not SWSSession[modName] then
			SWSSession[modName] = {}
		end
		local oldValue = SWSSession[modName][name]
		-- this way we ignore the first received data from someone and update afterwards
		-- this mimics Violation mechanics better
		if oldValue and oldValue < value then
			runningModules[modName]:OnRemoteData(sender, value-oldValue)
		end
		SWSSession[modName][sender] = value
	end
end

local function HandleSWSMessage( sender, msg )
	local type, name, data = select( 3, msg:find( SWSSplitMessage ) )
	if type and type == "BUDD" or type == "BUDR" then
		local i = 0
		local numbers = new()
		for tmp in data:gmatch( "%d+" ) do
			i = i + 1
			numbers[i] = tonumber(tmp)
		end
		numbers[1] = numbers[1] + numbers[10]			-- damage + dotdamage
		numbers[6] = numbers[6] + numbers[8]			-- overhealing + overhealinginfight
		numbers[2] = numbers[2] + numbers[7] - numbers[6]	-- healing + healinginfight - overhealing 
		local mod
		for k, value in ipairs( numbers ) do
			mod = type == "BUDD" and SWSmodulesOUT[k] or SWSmodulesIN[k]
			if mod and runningModules[mod] then
				AddSWSData( sender, name, mod, value )
			end
		end
		numbers = del(numbers)
	end
end

-- General SWS message handler
function Violation.commSpies:SWSudX21(message, distribution, sender)
	if not db.spying.swStats then
		return
	end
	local session, data = select( 3, message:find(SWSSyncMessage) )
	if not session or not data then
		return
	end
	for msg in data:gmatch( SWSSplitSync ) do
		HandleSWSMessage( sender, msg )
	end
end

-- SWS Handshake handler, we forward this to the main handler, since we handle everything there anyway
function Violation.commSpies:SWSHSX21(message, distribution, sender)
	local spy = self.commSpies["SWSudX21"]
	spy(self, message, distribution, sender )
end



----------------------------
-- GUI STUFF
----------------------------

---------------------------
-- Locals
---------------------------

local cachedBars = {}
local windowCache = {}

Violation.windows = {}

---------------------------
-- Initialization from Violation
---------------------------

function Violation:SetupWindows()
	for i = 1, self.db.char.windows.count do
		if not self.windows[i] then
			self.windows[i] = self:CreateWindow( self.db.char.windows[i] )
		end
	end
end

function Violation:CreateWindow( options )
	local w 
	if #windowCache > 0 then
		w = table.remove(windowCache) 
	else
		w = self.VioWindow:new()
	end
	w:Setup( options )
	return w
end

function Violation:DeleteWindow( window )
	local oldindex
	local wdb = self.db.char.windows
	for i = 1, wdb.count do
		if self.windows[i] == window then
			oldindex = i
		end
	end
	for i = oldindex, wdb.count - 1 do
		wdb[i] = wdb[i+1]
		self.windows[i] = self.windows[i+1]
	end
	table_insert( windowCache, window )
	wdb[wdb.count] = nil
	self.windows[wdb.count] = nil
	wdb.count = wdb.count - 1
	return nil
end

function Violation:UpdateWindows( sameamount )
	for k, w in ipairs( self.windows ) do
		w:Update( sameamount )
	end
end

function Violation:UpdateDirtyWindows()
	local dirtpile = new()
	for k, w in ipairs( self.windows ) do
		local modName = w:UpdateIfDirty()
		if modName then
			table_insert( dirtpile, modName )
		end
	end
	for k, modName in ipairs( dirtpile ) do
		local mod = self:GetModule( modName )
		mod:ResetDirty()
	end
	dirtpile = del(dirtpile)
end

----------------------------
-- Bars Local Helpers
----------------------------


local function NewBar( texture )
	local f
	if #cachedBars > 0 then
		f = table.remove(cachedBars)
	else
		f = CreateFrame("Button", nil, UIParent)
	end
	f:Hide()
	f:SetParent( UIParent )
	
	-- f:SetWidth( 150 )
	f:SetHeight( 15 )

	if not f.statusbar then
		f.statusbar = CreateFrame("StatusBar", nil, f)
	end
	f.statusbar:ClearAllPoints()
	f.statusbar:SetAllPoints( f ) 
	f.statusbar:SetStatusBarTexture(SM:Fetch("statusbar", texture) )
	f.statusbar:SetStatusBarColor(.5, .5, .5, 1)
	f.statusbar:SetMinMaxValues(0,100)
	f.statusbar:SetValue(0)

	if not f.lefttext then
		f.lefttext = f.statusbar:CreateFontString(nil, "OVERLAY")
	end
	f.lefttext:SetFontObject(GameFontHighlightSmall)
	f.lefttext:SetPoint("LEFT", f.statusbar, "LEFT", 0, 0)
	f.lefttext:SetJustifyH("LEFT")
	f.lefttext:SetText("")
	f.lefttext:SetTextColor(1,1,1,1)

	if not f.righttext then
		f.righttext = f.statusbar:CreateFontString(nil, "OVERLAY")
	end
	f.righttext:SetFontObject(GameFontHighlightSmall)
	f.righttext:SetPoint("RIGHT", f.statusbar, "RIGHT", 0, 0)
	f.righttext:SetJustifyH("RIGHT")
	f.righttext:SetText("")
	f.righttext:SetTextColor(1, 1, 1, 1)

	return f
end

local function FreeBar( f )
	f:Hide()
	f:SetParent( UIParent )
	f:ClearAllPoints()
	table_insert( cachedBars, f )
	return nil
end

----------------------------
-- Vio Window Local Helpers
----------------------------

local function OnWindowMouseUp( this, button )
	if button == "RightButton" then
		-- dew:Open( f, "children", function() dew:FeedAceOptionsTable(this.window.menuoptions) end, "cursorX", true, "cursorY", true )
	end
end

local function OnWindowMouseWheel( this, dir )
	this.window:OnOffset(dir > 0)
end

local function OnWindowClose( this )
	this:GetParent().window:Delete()
-- FIXME: just Close the window and provide an option to open it again
--	this:GetParent().window.options.shown = false
--	this:GetParent().window:Close()
end

local function OnWindowPrevModule( this )
	this:GetParent().window:SwitchView( false )
end

local function OnWindowNextModule( this )
	this:GetParent().window:SwitchView( true )
end

local function OnWindowDragStop( this )
	this:StopMovingOrSizing()
	local s = this:GetEffectiveScale()
	this.window.options.posX = this:GetLeft() * s
	this.window.options.posY = this:GetTop() * s
end

local function OnWindowSizeChanged( this )
	this.window:OnSized()
end

local function OnDragHandleShow( this )
	this.draghandle:Show()
end

local function OnDragHandleHide( this )
	this.draghandle:Hide()
end

local function ScheduleDragHandleHide( this )
	if this.windowframe then
		this = this.windowframe
	end
	if this.hideEvent then
		Violation:CancelScheduledEvent( this.hideEvent )
		this.hideEvent = nil
	end
	this.hideEvent = Violation:ScheduleEvent(OnDragHandleHide, 1, this)
end

local function OnDragHandleMouseDown( this )
	this.windowframe.isResizing = true
	this.windowframe:StartSizing("BOTTOMRIGHT")
end

local function OnDragHandleMouseUp( this, button )
	this.windowframe:StopMovingOrSizing()
	this.windowframe.window:OnSized()
	this.windowframe.isResizing = nil
end

local function OnDragHandleEnter( this )
	local hideEvent = this.hideEvent
	this.hideEvent = nil
	if this.windowframe then
		hideEvent = this.windowframe.hideEvent
		this.windowframe.hideEvent = nil
	end
	Violation:CancelScheduledEvent( hideEvent )
end

----------------------------
-- Violation Window Class 
----------------------------

local VioWindow = AceOO.Class()

Violation.VioWindow = VioWindow

function VioWindow.prototype:init()
	VioWindow.super.prototype.init(self)

	local f = CreateFrame("Frame", nil, UIParent)
	f:Hide()
	f.window = self
	self.frame = f
end

function VioWindow.prototype:Setup( options )
	self.options = options
	local f = self.frame
	local o = options
	local font = GameFontNormal:GetFont()

	self.barOffset = 0
	if not self.currentBars then
		self.currentBars = {}
	end
	
	self:SetupMenuOptions()
	
	f:SetScale(o.scale or 1)
	f:SetClampedToScreen(true)

	f:SetWidth( o.width )
	f:SetHeight( o.height )
	f:SetPoint( "CENTER", UIParent, "CENTER")
	f:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface\\AddOns\\Violation\\textures\\otravi-semi-full-border", edgeSize = 32,
		insets = {left = 1, right = 1, top = 20, bottom = 1},
	})
	f:SetBackdropColor(24/255, 24/255, 24/255)
	f:EnableMouse(true)
	f:EnableMouseWheel(true)

	f:SetScript("OnMouseUp", OnWindowMouseUp )
	f:SetScript("OnMouseWheel", OnWindowMouseWheel ) 

	-- header
	if not f.header then
		f.header = f:CreateFontString(nil,"OVERLAY")
	end
	f.header:SetFont(font, 12)
	f.header:SetText("")

	local module = Violation:HasModule(options.module) and Violation:GetModule(options.module) or nil

	if module then f.header:SetText( module:GetDisplayName() ) end
	f.header:SetJustifyH("LEFT")
	f.header:SetShadowOffset(.8, -.8)
	f.header:SetShadowColor(0, 0, 0, 1)
	f.header:ClearAllPoints()
	f.header:SetPoint("TOPLEFT", f, "TOPLEFT", 24, -14)
	f.header:SetPoint("TOPRIGHT", f, "TOPRIGHT", -42, -14)

	-- close button
	if not f.closebutton then
		f.closebutton = CreateFrame("Button", nil, f)
	end
	f.closebutton:SetWidth(16)
	f.closebutton:SetHeight(16)
	f.closebutton:SetPoint("TOPRIGHT", f, "TOPRIGHT", -7, -13)
	f.closebutton:SetScript( "OnClick", OnWindowClose )
	if not f.close then
		f.close = f:CreateTexture(nil, "ARTWORK")
	end
	f.close:SetPoint("CENTER", f.closebutton, "CENTER")
	f.close:SetTexture("Interface\\AddOns\\Violation\\textures\\otravi-close")
	f.close:SetWidth(16)
	f.close:SetHeight(16)
	f.close:SetBlendMode("ADD")

	-- prev button
	if not f.prevbutton then
		f.prevbutton = CreateFrame("Button", nil, f)
	end
	f.prevbutton:SetWidth(16)
	f.prevbutton:SetHeight(16)
	f.prevbutton:SetPoint("TOPLEFT", f, "TOPLEFT", 7, -13)
	f.prevbutton:SetScript( "OnClick", OnWindowPrevModule )
	if not f.prev then
		f.prev = f:CreateTexture(nil, "ARTWORK")
	end
	f.prev:SetPoint("CENTER", f.prevbutton, "CENTER")
	f.prev:SetTexture("Interface\\AddOns\\Violation\\textures\\prev")
	f.prev:SetWidth(16)
	f.prev:SetHeight(16)
	f.prev:SetBlendMode("ADD")
	
	-- next button
	if not f.nextbutton then
		f.nextbutton = CreateFrame("Button", nil, f)
	end
	f.nextbutton:SetWidth(16)
	f.nextbutton:SetHeight(16)
	f.nextbutton:SetPoint("TOPRIGHT", f, "TOPRIGHT", -28, -13)
	f.nextbutton:SetScript( "OnClick", OnWindowNextModule )
	if not f.next then
		f.next = f:CreateTexture(nil, "ARTWORK")
	end
	f.next:SetPoint("CENTER", f.nextbutton, "CENTER")
	f.next:SetTexture("Interface\\AddOns\\Violation\\textures\\next")
	f.next:SetWidth(16)
	f.next:SetHeight(16)
	f.next:SetBlendMode("ADD")
	
	-- drag handle
	if not f.draghandle then
		f.draghandle = CreateFrame("Frame", nil, f)
	end
	f.draghandle.windowframe = f
	f.draghandle:Hide()
	f.draghandle:SetFrameLevel( f:GetFrameLevel() + 10 ) -- place this above everything
	f.draghandle:SetWidth(16)
	f.draghandle:SetHeight(16)
	f.draghandle:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -1, 1)
	f.draghandle:EnableMouse(true)

	if not f.draghandle.texture then
		f.draghandle.texture = f.draghandle:CreateTexture(nil,"BACKGROUND")
	end
	f.draghandle.texture:SetTexture("Interface\\AddOns\\Violation\\textures\\draghandle")
	f.draghandle.texture:SetWidth(16)
	f.draghandle.texture:SetHeight(16)
	f.draghandle.texture:SetBlendMode("ADD")
	f.draghandle.texture:SetPoint( "CENTER", f.draghandle, "CENTER", 0, 0)

	self:ToggleLock( o.locked )
	self:RestorePosition()
	
	self:Update()
	
	if o.shown then
		f:Show()
	end
end

function VioWindow.prototype:SetupMenuOptions()
end

function VioWindow.prototype:ToggleLock( locked )
	self.options.locked = locked
	local f = self.frame
	if locked then
		f:SetMovable(nil)
		f:SetResizable(nil)
		f:RegisterForDrag()
		f.closebutton:Hide()
		f.close:Hide()

		f:SetScript("OnEnter", nil)
		f:SetScript("OnLeave", nil)
		f:SetScript("OnDragStart", nil)
		f:SetScript("OnDragStop", nil)
		f:SetScript("OnSizeChanged", nil)
		f.draghandle:SetScript("OnMouseDown", nil)
		f.draghandle:SetScript("OnMouseUp", nil)
		f.draghandle:SetScript("OnEnter", nil)
		f.draghandle:SetScript("OnLeave", nil)
	else
		f:SetMovable(true)
		f:SetResizable(true)
		f:SetMinResize(50,50)
		f:RegisterForDrag("LeftButton")
		f.closebutton:Show()
		f.close:Show()

		f:SetScript("OnDragStart", f.StartMoving)
		f:SetScript("OnDragStop", OnWindowDragStop)
		f:SetScript("OnSizeChanged", OnWindowSizeChanged)
		f:SetScript("OnEnter", OnDragHandleShow)
		f:SetScript("OnLeave", ScheduleDragHandleHide )
		f.draghandle:SetScript("OnMouseDown", OnDragHandleMouseDown)
		f.draghandle:SetScript("OnMouseUp", OnDragHandleMouseUp)
		f.draghandle:SetScript("OnEnter", OnDragHandleEnter)
		f.draghandle:SetScript("OnLeave", ScheduleDragHandleHide)
	end

end

function VioWindow.prototype:RestorePosition()
	local f = self.frame
	local x = self.options.posX
	local y = self.options.posY
	if x and y then
		local s = f:GetEffectiveScale()
		f:ClearAllPoints()
		f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x / s, y / s)
	end
end

function VioWindow.prototype:OnOffset( up )
	-- we do not check if ouyr barOffset is still valid here. Update will do that for us
	if up then
		self.barOffset = self.barOffset - 1
	else
		self.barOffset = self.barOffset + 1
	end
	self:Update()
end

function VioWindow.prototype:OnSized()
	local f = self.frame
	local o = self.options

	if not f.isResizing then return end -- don't do anything on frame creation

	local y = f:GetHeight()
	local x = f:GetWidth()

	o.height = y
	o.width = x

	y = math.ceil(y)
	y = y - 34 -- deduct the header size

	local nrbars = math.floor( y / 16 )
	if nrbars < 1 then nrbars = 1 end

	if nrbars ~= o.nrBars then
		if nrbars < o.nrBars then
			for i = nrbars+1, o.nrBars do
				self:ClearBar( i )
			end
		end
		o.nrBars = nrbars
		self:Update()
	else
		self:Update(true)
	end
end

function VioWindow.prototype:SwitchView( forward )
	local o = self.options
	local mod = o.module
	local dirtpile = new()
	local first, last
	for name in Violation:IterateModules() do
		if Violation.runningModules[name] then
			table_insert( dirtpile, name )
			if not first then
				first = name
			end
			last = name
		end
	end
	if forward and last == mod then
		o.module = first
	elseif not forward and first == mod then
		o.module = last
	else
		local p
		for k, v in ipairs( dirtpile ) do
			if forward then
				if v == mod then
					p = v
				elseif p then
					o.module = v
					break
				end
			else
				if v == mod then
					o.module = p
					break
				else
					p = v
				end
			end
		end
	end
	dirtpile = del(dirtpile)
	if mod ~= o.module then
		self:Update()
	end
end

function VioWindow.prototype:Delete()
	self:Close()
	Violation:DeleteWindow( self )
end

function VioWindow.prototype:Close()
	self.frame:Hide()
	for k, v in ipairs( self.currentBars ) do
		self:ClearBar( k )
	end
end

function VioWindow.prototype:Open()
	self.frame:Show()
	self:Update()
end

function VioWindow.prototype:SetBar( count, rank, name, textValue, value, vmax, percent )
	if not self.currentBars[count] then
		self.currentBars[count] = NewBar( self.options.barTexture )
	end
	local f = self.currentBars[count]

	f:SetParent( self.frame )
	
	f.lefttext:SetText( format("%s %s", rank, name) )

	if percent then
		f.righttext:SetText( format("%s %s", textValue, percentString:format(percent)) )
	else
		f.righttext:SetText( textValue )
	end
	if vmax and vmax ~= 0 then
		f.statusbar:SetValue((value/vmax)*100)
	else
		f.statusbar:SetValue(100)
	end
	local color = RGB[name]
	if not color then
		f.statusbar:SetStatusBarColor( .5, .5, .5, 1)
	else
		f.statusbar:SetStatusBarColor( color[1], color[2], color[3], 1)
	end
	if count == 1 then
		f:SetPoint( "TOP", self.frame, "TOP", 0, -32 )
	else
		f:SetPoint( "TOP", self.currentBars[count-1], "BOTTOM", 0, -1 )
	end
	f.statusbar:SetStatusBarTexture(SM:Fetch("statusbar", self.options.barTexture) )
	f:SetPoint( "LEFT", self.frame, "LEFT", 2, 0)
	f:SetPoint( "RIGHT", self.frame, "RIGHT", -2, 0)
	f:Show()
end

function VioWindow.prototype:ClearBar( count )
	if not self.currentBars[count] then return end
	self.currentBars[count] = FreeBar( self.currentBars[count] )	
end

function VioWindow.prototype:UpdateIfDirty()
	local mod = Violation:HasModule(self.options.module) and Violation:GetModule(self.options.module) or nil
	if mod and mod:IsDirty() then
		self:Update()
		return self.options.module
	end
end

function VioWindow.prototype:Update( sameamount )
	if not self.options.shown then return end

	local o = self.options

	local mod = o.module
	local module = Violation:GetModule(mod)
	if not module then return end
	
	local f = self.frame
	
	if( (f.header:GetRight() or 0) - (f.header:GetLeft() or 0) < 80) then
		f.header:SetText( module:GetShortDisplayName() )
	else
		f.header:SetText( module:GetDisplayName() )
	end

	local color = o.useModuleColor and Violation.db.char.colors[mod] or o.color
	f:SetBackdropBorderColor( color.r, color.g, color.b )
	
	if f.barScale~=(o.scale or 1)  then
		local x = f:GetLeft()
		local y = f:GetTop()
		local scale = f:GetScale()
		x=x*scale
		y=y*scale
		scale = o.scale or 1
		f:SetScale(scale)
		x=x/scale
		y=y/scale
		f:ClearAllPoints()
		f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x,y)
		f.barScale = o.scale
	end
	
	if( not sameamount ) then
		local data = Violation:GetModuleData(module.name)
		if type(data) == "table" then 
			self:UpdateBars(module.name)
		end
	end
end

function VioWindow.prototype:UpdateBars( modName )
	local o = self.options
	local barCount, barOff = Violation:PublishModuleData(modName, self.BarAddCallBack, o.nrBars, self.barOffset, self)
	self.barOffset = barOff -- set our barOffset to the correct new value.
	if barCount < o.nrBars then 
		for i = barCount+1, o.nrBars do
			self:ClearBar( i )
		end
	end
end

function VioWindow.prototype:BarAddCallBack(i, rank, name, value, percentage, dataInfo)
	return self:SetBar(i, rank..".", name, dataInfo.formatter(value), value, dataInfo.maxValue, percentage)
end

